from django.db import models
from django.urls import reverse


class Genre(models.Model):
    name = models.CharField(max_length=200, help_text='Enter a book genre (e.g. Science Fiction)')

    def __str__(self):
        return self.name


class Book(models.Model):
    title = models.CharField(max_length=200)
    library = models.ForeignKey('loans.Library', related_name='books', on_delete=models.CASCADE)
    author = models.ForeignKey('Author', on_delete=models.SET_NULL, null=True)
    summary = models.TextField(max_length=1000, help_text='Enter a brief description of the book')
    isbn = models.CharField(
        'ISBN',
        max_length=13,
        help_text=(
            '13 Character '
            '<a href="https://www.isbn-international.org/content/what-isbn">'
            'ISBN number</a>',
        ),
    )
    genre = models.ManyToManyField(Genre, help_text='Select a genre for this book')
    published_on = models.DateField()
    last_print = models.DateField(auto_now_add=True)
    pages = models.IntegerField(null=True)

    def get_absolute_url(self):
        return reverse(
            'admin:books_book_change', args=(self.id,)
        )

    def __str__(self):
        return self.title


class Author(models.Model):
    first_name = models.CharField(max_length=100)
    last_name = models.CharField(max_length=100)
    date_of_birth = models.DateField(null=True, blank=True)
    date_of_death = models.DateField('Died', null=True, blank=True)

    class Meta:
        ordering = ('last_name', 'first_name')

    def get_absolute_url(self):
        return reverse(
            'admin:books_author_change', args=(self.id,)
        )

    def __str__(self):
        return '{}, {}'.format(self.first_name, self.last_name)



import re

import django
import pytest

from richmin.compat import reverse
from .test_app.library.books.models import Book
from .test_app.library.factories import BookFactory


@pytest.mark.django_db
def test_login(client, admin_user):
    """
    We can render the login page
    """
    url = reverse('admin:login')

    response = client.get(url)
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    assert templates_used == [
        'admin/login.html',
        'registration/base.html',
    ]

    response = client.post(
        url + '?next=/admin/',
        data={
            'username': admin_user.username,
            'password': 'password'
        },
        follow=True,
    )

    assert response.status_code == 200
    assert 'dashboard' in response.content.decode()


@pytest.mark.django_db
def test_logout(admin_client):
    """
    We can log out, and render the logout page
    """
    url = reverse('admin:logout')

    response = admin_client.get(url)
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    assert templates_used == [
        'registration/logged_out.html',
        'registration/base.html',
    ]


@pytest.mark.django_db
def test_reset_password(client, admin_user):
    """
    We can render the password reset views
    """
    # Step 1: Password reset form
    url = reverse('admin_password_reset')

    response = client.get(url)
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    assert templates_used == [
        'registration/password_reset_form.html',
        'registration/base.html',
    ]

    # Step 2: Password reset done
    response = client.post(
        url,
        data={'email': admin_user.email},
        follow=True,
    )
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    assert response.resolver_match.url_name == 'password_reset_done'
    assert templates_used == [
        'registration/password_reset_done.html',
        'registration/base.html',
    ]
    assert 'We’ve emailed you instructions for setting your password' in response.content.decode()

    # Get password reset link from reset email
    email = django.core.mail.outbox[0]
    url = re.search(r'https?://[^/]*(/.*reset/\S*)', email.body).group(1)

    # Step 3: Password reset confirm
    response = client.get(url, follow=True)
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    assert response.resolver_match.url_name == 'password_reset_confirm'
    assert templates_used == [
        'registration/password_reset_confirm.html',
        'registration/base.html',
    ]

    # Step 4: Password reset complete
    response = client.post(
        response.request['PATH_INFO'],
        data={
            'username': admin_user.username,
            'new_password1': 'new_password',
            'new_password2': 'new_password'
        },
        follow=True,
    )
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    assert response.resolver_match.url_name == 'password_reset_complete'
    assert templates_used == [
        'registration/password_reset_complete.html',
        'registration/base.html',
    ]
    assert 'Your password has been set.' in response.content.decode()


@pytest.mark.django_db
def test_password_change(admin_client):
    """
    We can render the password change form, and successfully change our password
    """
    url = reverse('admin:password_change')

    response = admin_client.get(url)
    templates_used = [t.name for t in response.templates]
    expected_templates_used = {
        'registration/password_change_form.html',
        'admin/base_site.html',
        'admin/base.html',
        'django/forms/widgets/password.html',
        'django/forms/widgets/input.html',
        'django/forms/widgets/attrs.html',
        'django/forms/widgets/password.html',
        'django/forms/widgets/input.html',
        'django/forms/widgets/attrs.html',
        'django/forms/widgets/password.html',
        'django/forms/widgets/input.html',
        'django/forms/widgets/attrs.html',
        'richmin/includes/ui_builder_panel.html',
    }

    if django.VERSION[0] == 4:
        expected_templates_used.update({'django/forms/errors/list/default.html', 'django/forms/errors/list/ul.html'})

    assert response.status_code == 200
    assert set(templates_used) == expected_templates_used

    response = admin_client.post(
        url,
        data={
            'old_password': 'password',
            'new_password1': 'PickleRick123!!',
            'new_password2': 'PickleRick123!!'
        },
        follow=True,
    )
    templates_used = [t.name for t in response.templates]

    assert 'Password change successful' in response.content.decode()
    assert set(templates_used) == {
        'registration/password_change_done.html',
        'admin/base.html',
        'admin/base_site.html',
        'richmin/includes/ui_builder_panel.html',
    }


@pytest.mark.django_db
def test_dashboard(admin_client):
    """
    We can render the dashboard
    """
    url = reverse('admin:index')

    response = admin_client.get(url)
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    assert templates_used == [
        'admin/index.html',
        'admin/base_site.html',
        'admin/base.html',
        'richmin/includes/ui_builder_panel.html',
    ]


@pytest.mark.django_db
def test_detail(admin_client):
    """
    We can render the detail view
    """
    book = BookFactory()
    url = reverse(
        'admin:books_book_change', args=(book.pk,)
    )

    response = admin_client.get(url)
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    render_counts = {x: templates_used.count(x) for x in set(templates_used)}

    expected_render_counts = {
        'admin/base.html': 1,
        'admin/base_site.html': 1,
        'admin/change_form.html': 1,
        'admin/change_form_object_tools.html': 1,
        'admin/edit_inline/stacked.html': 1,
        'admin/includes/fieldset.html': 4,
        'admin/prepopulated_fields_js.html': 1,
        'admin/submit_line.html': 1,
        'admin/widgets/foreign_key_raw_id.html': 1,
        'admin/widgets/related_widget_wrapper.html': 4,
        'admin/widgets/split_datetime.html': 2,
        'django/forms/widgets/attrs.html': 47,
        'django/forms/widgets/date.html': 5,
        'django/forms/widgets/hidden.html': 6,
        'django/forms/widgets/input.html': 19,
        'django/forms/widgets/number.html': 1,
        'django/forms/widgets/select.html': 6,
        'django/forms/widgets/select_option.html': 21,
        'django/forms/widgets/text.html': 4,
        'django/forms/widgets/textarea.html': 1,
        'django/forms/widgets/time.html': 2,
        'richmin/includes/horizontal_tabs.html': 1,
        'richmin/includes/ui_builder_panel.html': 1,
    }

    if django.VERSION[0] == 4:
        expected_render_counts.update({
            'django/forms/div.html': 1,
            'django/forms/errors/list/default.html': 2,
            'django/forms/errors/list/ul.html': 56,
        })

    # The number of times each template was rendered
    assert render_counts == expected_render_counts

    expected_templates_used = {
        'admin/base.html',
        'admin/base_site.html',
        'admin/change_form.html',
        'admin/change_form_object_tools.html',
        'admin/edit_inline/stacked.html',
        'admin/includes/fieldset.html',
        'admin/prepopulated_fields_js.html',
        'admin/submit_line.html',
        'admin/widgets/foreign_key_raw_id.html',
        'admin/widgets/related_widget_wrapper.html',
        'admin/widgets/split_datetime.html',
        'django/forms/widgets/attrs.html',
        'django/forms/widgets/date.html',
        'django/forms/widgets/hidden.html',
        'django/forms/widgets/input.html',
        'django/forms/widgets/number.html',
        'django/forms/widgets/select.html',
        'django/forms/widgets/select_option.html',
        'django/forms/widgets/text.html',
        'django/forms/widgets/textarea.html',
        'django/forms/widgets/time.html',
        'richmin/includes/horizontal_tabs.html',
        'richmin/includes/ui_builder_panel.html',
    }

    if django.VERSION[0] == 4:
        expected_templates_used.update({
            'django/forms/div.html',
            'django/forms/errors/list/default.html',
            'django/forms/errors/list/ul.html',
        })

    # The templates that were used
    assert set(templates_used) == expected_templates_used

    # TODO: post data and confirm we can change model instances


@pytest.mark.django_db
def test_list(admin_client):
    """
    We can render the list view
    """
    BookFactory.create_batch(5)

    url = reverse('admin:books_book_changelist')

    response = admin_client.get(url)
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    render_counts = {x: templates_used.count(x) for x in set(templates_used)}

    expected_render_counts = {
        'admin/actions.html': 2,
        'admin/base.html': 1,
        'admin/base_site.html': 1,
        'admin/change_list.html': 1,
        'admin/change_list_object_tools.html': 1,
        'admin/change_list_results.html': 1,
        'admin/date_hierarchy.html': 1,
        'admin/pagination.html': 1,
        'admin/search_form.html': 1,
        'django/forms/widgets/attrs.html': 27,
        'django/forms/widgets/checkbox.html': 5,
        'django/forms/widgets/hidden.html': 11,
        'django/forms/widgets/input.html': 21,
        'django/forms/widgets/select.html': 2,
        'django/forms/widgets/select_option.html': 4,
        'django/forms/widgets/text.html': 5,
        'richmin/includes/ui_builder_panel.html': 1,
    }

    if django.VERSION[0] == 4:
        expected_render_counts.update({
            'django/forms/div.html': 1,
            'django/forms/errors/list/default.html': 6,
            'django/forms/errors/list/ul.html': 6,
        })

    # The number of times each template was rendered
    assert render_counts == expected_render_counts

    expected_templates = {
        'admin/actions.html',
        'admin/base.html',
        'admin/base_site.html',
        'admin/change_list.html',
        'admin/change_list_object_tools.html',
        'admin/change_list_results.html',
        'admin/date_hierarchy.html',
        'admin/pagination.html',
        'admin/search_form.html',
        'django/forms/widgets/attrs.html',
        'django/forms/widgets/checkbox.html',
        'django/forms/widgets/hidden.html',
        'django/forms/widgets/input.html',
        'django/forms/widgets/select.html',
        'django/forms/widgets/select_option.html',
        'django/forms/widgets/text.html',
        'richmin/includes/ui_builder_panel.html',
    }

    if django.VERSION[0] == 4:
        expected_templates.update({
            'django/forms/div.html',
            'django/forms/errors/list/default.html',
            'django/forms/errors/list/ul.html',
        })

    # The templates that were used
    assert set(templates_used) == expected_templates


@pytest.mark.django_db
def test_history(admin_client):
    """
    We can render the object history page
    """
    book = BookFactory()
    url = reverse(
        'admin:books_book_history', args=(book.pk,)
    )

    response = admin_client.get(url)
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    render_counts = {x: templates_used.count(x) for x in set(templates_used)}

    # The number of times each template was rendered
    assert render_counts == {
        'admin/object_history.html': 1,
        'admin/base.html': 1,
        'admin/base_site.html': 1,
        'richmin/includes/ui_builder_panel.html': 1,
    }

    # The templates that were used
    assert set(templates_used) == {
        'admin/object_history.html',
        'admin/base.html',
        'admin/base_site.html',
        'richmin/includes/ui_builder_panel.html',
    }


@pytest.mark.django_db
def test_delete(admin_client):
    """
    We can load the confirm delete page, and POST it, and it deletes our object
    """
    book = BookFactory()
    url = reverse(
        'admin:books_book_delete', args=(book.pk,)
    )

    response = admin_client.get(url)
    templates_used = [t.name for t in response.templates]

    assert response.status_code == 200
    render_counts = {x: templates_used.count(x) for x in set(templates_used)}

    # The number of times each template was rendered
    assert render_counts == {
        'admin/delete_confirmation.html': 1,
        'admin/base_site.html': 1,
        'admin/base.html': 1,
        'admin/includes/object_delete_summary.html': 1,
        'richmin/includes/ui_builder_panel.html': 1,
    }

    # The templates that were used
    assert set(templates_used) == {
        'admin/delete_confirmation.html',
        'admin/base_site.html',
        'admin/base.html',
        'admin/includes/object_delete_summary.html',
        'richmin/includes/ui_builder_panel.html',
    }

    response = admin_client.post(url, data={'post': 'yes'}, follow=True)

    # We deleted our object, and are now back on the changelist
    assert not Book.objects.all().exists()
    assert response.resolver_match.url_name == 'books_book_changelist'



# Generated by Django 2.2.15 on 2020-10-15 11:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("loans", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="library",
            options={"verbose_name_plural": "Libraries"},
        ),
        migrations.AddField(
            model_name="library",
            name="name",
            field=models.CharField(default=1, max_length=255),
            preserve_default=False,
        ),
    ]



from unittest.mock import MagicMock, Mock, patch

import pytest
from django.db.models.functions import Upper
from django.urls import reverse

from richmin.utils import get_admin_url, get_app_admin_urls, get_custom_url, get_model_meta, get_view_permissions, \
    order_with_respect_to
from .test_app.library.books.models import Book
from .test_app.library.factories import BookFactory, UserFactory


def test_order_with_respect_to():
    """
    When we ask for ordering, we get it as expected
    """

    def apps(*args):
        return [{'app_label': x} for x in args]

    original_list = apps('b', 'c', 'a')

    assert order_with_respect_to(original_list, ['c', 'b'], getter=lambda x: x['app_label']) == apps('c', 'b', 'a')
    assert order_with_respect_to(original_list, ['nothing'], getter=lambda x: x['app_label']) == original_list
    assert order_with_respect_to(original_list, ['a'], getter=lambda x: x['app_label'])[0]['app_label'] == 'a'
    assert order_with_respect_to([1, 2, 3], [3, 2, 1]) == [3, 2, 1]
    assert order_with_respect_to([1, 2, 3], [3]) == [3, 1, 2]
    assert order_with_respect_to(['morty', 'pickle', 'rick'], ['pickle', 'morty']) == [
        'pickle',
        'morty',
        'rick',
    ]


@pytest.mark.django_db
def test_get_admin_url(admin_user):
    """
    We can get admin urls for Model classes, instances, or app.model strings
    """
    book = BookFactory()

    assert get_admin_url(book) == reverse(
        'admin:books_book_change', args=(book.pk,)
    )
    assert get_admin_url(Book) == reverse('admin:books_book_changelist')
    assert get_admin_url(Book, q='test') == reverse('admin:books_book_changelist') + '?q=test'
    assert get_admin_url('books.Book') == reverse('admin:books_book_changelist')
    with patch('richmin.utils.reverse') as mock_reverse:
        get_admin_url('Books.Book')
        mock_reverse.assert_called_once_with('admin:Books_book_changelist', current_app='admin')
    assert get_admin_url('cheese:bad_pattern') == '#'
    assert get_admin_url('fake_app.fake_model') == '#'
    assert get_admin_url(1) == '#'


def test_get_custom_url():
    """
    We handle urls that can be reversed, and that cant, and external links
    """
    assert get_custom_url('http://somedomain.com') == 'http://somedomain.com'
    assert get_custom_url('/relative/path') == '/relative/path'
    assert get_custom_url('admin:books_book_changelist') == '/en/admin/books/book/'


@pytest.mark.django_db
def test_get_model_meta(admin_user):
    """
    We can fetch model meta
    """
    assert get_model_meta('auth.user') == admin_user._meta
    assert get_model_meta('books.book') == Book._meta
    assert get_model_meta('nothing') is None
    assert get_model_meta('nothing.nothing') is None


@pytest.mark.django_db
def test_get_app_admin_urls():
    """
    We can get all the admin urls for an app
    """
    assert get_app_admin_urls('books') == [
        {
            'url': '/en/admin/books/genre/',
            'model': 'books.genre',
            'name': 'Genres'
        },
        {
            'url': '/en/admin/books/book/',
            'model': 'books.book',
            'name': 'Books'
        },
        {
            'url': '/en/admin/books/author/',
            'model': 'books.author',
            'name': 'Authors'
        },
    ]

    assert get_app_admin_urls('nothing') == []


@pytest.mark.django_db
def test_get_model_permissions():
    """
    We can create the correct model permissions from user permissions
    """
    user = UserFactory(permissions=('books.view_book', 'books.view_author'))

    assert get_view_permissions(user) == {'books.book', 'books.author'}

    # test for camel cased app names
    user = MagicMock()
    user.get_all_permissions = Mock(return_value={'BookShelf.view_author', 'BookShelf.view_book'})

    assert get_view_permissions(user) == {'BookShelf.book', 'BookShelf.author'}


@pytest.mark.django_db
def test_get_model_permissions_lowercased():
    """
    When our permissions are upper cased (we had an app with an upper case letter) we still get user perms in lower case
    """
    user = UserFactory(permissions=('books.view_book', 'books.view_author'))
    user.user_permissions.update(codename=Upper('codename'))

    assert get_view_permissions(user) == {'books.book', 'books.author'}



# Generated by Django 3.1.5 on 2021-01-18 15:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("books", "0003_auto_20201015_1222"),
    ]

    operations = [
        migrations.AddField(
            model_name="book",
            name="pages",
            field=models.IntegerField(null=True),
        ),
    ]



# Generated by Django 2.2.16 on 2021-02-26 21:49

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("books", "0004_book_pages"),
    ]

    operations = [
        migrations.AddField(
            model_name="book",
            name="last_print",
            field=models.DateField(default=datetime.date(2000, 1, 1)),
            preserve_default=False,
        ),
    ]



from django.contrib import admin
from django.urls import path

from .models import BookLoan, Library
from .views import CustomView


class BookLoanInline(admin.StackedInline):
    model = BookLoan
    extra = 1
    readonly_fields = ('id', 'duration')
    fields = (
        'book',
        'imprint',
        'status',
        'due_back',
        'borrower',
        'loan_start',
        'duration',
    )


@admin.register(BookLoan)
class BookLoanAdmin(admin.ModelAdmin):
    list_display = ('book', 'status', 'borrower', 'due_back', 'id')
    list_filter = ('status', 'due_back')
    autocomplete_fields = ('borrower',)
    search_fields = ('book__title',)
    readonly_fields = ('id',)
    fieldsets = (
        (None, {
            'fields': ('book', 'imprint', 'id')
        }),
        ('Availability', {
            'fields': ('status', 'due_back', 'duration', 'borrower')
        }),
    )

    def get_urls(self):
        """
        Add in a custom view to demonstrate =
        """
        urls = super().get_urls()
        return urls + [path('custom_view', CustomView.as_view(), name='custom_view')]

    def response_change(self, request, obj):
        ret = super().response_change(request, obj)

        if 'reserve' in request.POST:
            obj.status = 'r'
            obj.save()
        return ret


@admin.register(Library)
class LibraryAdmin(admin.ModelAdmin):
    list_display = ('name', 'address', 'librarian')



from django.apps import AppConfig
from django.conf import settings


class BooksConfig(AppConfig):
    name = '{}library.books'.format(settings.PREFIX)

    def ready(self):
        from . import receivers  # NOQA



from datetime import date, timedelta

import factory
from django.contrib.auth.models import Group, Permission, User
from django.utils import timezone
from factory.django import DjangoModelFactory
from factory.fuzzy import FuzzyChoice, FuzzyDate, FuzzyInteger

from .books.models import Author, Book, Genre
from .loans.models import BookLoan, Library

NOW = timezone.now()


class GroupFactory(DjangoModelFactory):
    name = factory.Faker('job')

    @factory.post_generation
    def permissions(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for group in extracted:
                self.permissions.add(group)

    class Meta:
        model = Group


class UserFactory(DjangoModelFactory):
    username = factory.Faker('name')
    email = factory.Faker('email')
    is_staff = True
    is_active = True
    is_superuser = False
    password = factory.PostGenerationMethodCall('set_password', 'test')

    @factory.post_generation
    def groups(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for group in extracted:
                self.groups.add(group)

    @factory.post_generation
    def permissions(self, create, extracted, **kwargs):
        """
        Create a user with the given permissions, e.g UserFactory(permissions=('books.view_book', 'auth.change_user'))
        """
        if not create:
            return

        if extracted:
            available_permissions = [
                '{}.{}'.format(x[0], x[1])
                for x in Permission.objects.values_list('content_type__app_label', 'codename')
            ]

            for permission in extracted:
                assert permission in available_permissions, '{} not in {}'.format(permission, available_permissions)

                app, perm = permission.split('.')
                perm_obj = Permission.objects.get(content_type__app_label=app, codename=perm)

                self.user_permissions.add(perm_obj)

    class Meta:
        model = User


class GenreFactory(DjangoModelFactory):
    name = factory.Faker('job')

    class Meta:
        model = Genre


class AuthorFactory(DjangoModelFactory):
    first_name = factory.Faker('first_name')
    last_name = factory.Faker('last_name')
    date_of_birth = FuzzyDate(date(1950, 1, 1), NOW.date() - timedelta(days=365))
    date_of_death = factory.LazyAttribute(
        lambda x: x.date_of_birth.replace(year=NOW.year, day=min(28, x.date_of_birth.day))
    )

    class Meta:
        model = Author


class LibraryFactory(DjangoModelFactory):
    name = factory.Faker('company')
    address = factory.Faker('address')
    librarian = factory.SubFactory(UserFactory)

    class Meta:
        model = Library


class BookFactory(DjangoModelFactory):
    title = factory.Faker('sentence')
    author = factory.SubFactory(AuthorFactory)
    library = factory.SubFactory(LibraryFactory)
    summary = factory.Faker('sentence')
    isbn = '9780123456472'
    published_on = FuzzyDate(date(1950, 1, 1), date(1999, 1, 1))
    last_print = FuzzyDate(date(2000, 1, 1), date(2020, 1, 1))
    pages = FuzzyInteger(50, 1000)

    @factory.post_generation
    def genre(self, create, extracted, **kwargs):
        if not create:
            return

        extracted = extracted or GenreFactory.create_batch(3)
        for genre in extracted:
            self.genre.add(genre)

    class Meta:
        model = Book


class BookLoanFactory(DjangoModelFactory):
    id = factory.Faker('uuid4')
    book = factory.SubFactory(BookFactory)
    imprint = factory.Faker('name')
    loan_start = NOW
    due_back = factory.LazyAttribute(lambda x: x.loan_start + timedelta(weeks=2))
    borrower = factory.SubFactory(UserFactory)
    status = FuzzyChoice(dict(BookLoan.LOAN_STATUS).keys())

    class Meta:
        model = BookLoan



from django.contrib.admin.sites import site
from django.views.generic import TemplateView


class CustomView(TemplateView):
    template_name = 'loans/custom.html'

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx.update(site.each_context(self.request))
        return ctx
